"""Custom Tensorflow callbacks."""

import os
import numpy as np
import tensorflow as tf
from tensorflow import tanh
from tensorflow.keras import callbacks

from osl_dynamics import inference


class DiceCoefficientCallback(callbacks.Callback):
    """Callback to calculate a Dice coefficient during training.

    Parameters
    ----------
    prediction_dataset : tf.data.Dataset
        Dataset to use to calculate outputs of the model.
    ground_truth_time_course : np.ndarray
        2D or 3D numpy array containing the ground truth state/mode time
        course of the training data. Shape must be (n_time_courses, n_samples,
        n_modes) or (n_samples, n_modes).
    names : list of str, optional
        Names for the time courses. Shape must be (n_time_courses,).
    """

    def __init__(
        self,
        prediction_dataset,
        ground_truth_time_course,
        names=None,
    ):
        super().__init__()
        self.prediction_dataset = prediction_dataset
        if ground_truth_time_course.ndim == 2:
            # We're training a single time scale model
            self.n_time_courses = 1
            self.gttc = ground_truth_time_course[np.newaxis, ...]
        elif ground_truth_time_course.ndim == 3:
            # We're training a multi-time-scale model
            self.n_time_courses = ground_truth_time_course.shape[0]
            self.gttc = ground_truth_time_course
        else:
            raise ValueError(
                "A 2D or 3D numpy array must be pass for ground_truth_time_course."
            )
        if names is not None:
            if len(names) != self.n_time_courses:
                raise ValueError(
                    "Mismatch between the number of names and time courses."
                )
        self.names = names
        self.n_modes = ground_truth_time_course.shape[-1]

    def on_epoch_end(self, epoch, logs=None):
        """Action to perform at the end of an epoch.

        Parameters
        ---------
        epochs : int
            Integer, index of epoch.
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """

        # Predict time courses
        predictions = self.model.predict(self.prediction_dataset, verbose=0)
        tc = predictions[2:]  # first two outputs are losses, rest are time courses
        if len(tc) != self.n_time_courses:
            raise ValueError(
                "Mismatch between number of ground truth and predicted time courses."
            )

        # For each time course calculate the dice with respect to the
        # ground truth
        dices = []
        for i in range(self.n_time_courses):
            pmtc = inference.modes.argmax_time_courses(
                tc[i], concatenate=True, n_modes=self.n_modes
            )
            pmtc, gttc = inference.modes.match_modes(pmtc, self.gttc[i])
            dice = inference.metrics.dice_coefficient(pmtc, gttc)
            dices.append(dice)

        # Add dice to the training history and print to screen
        if self.n_time_courses == 1:
            logs["dice"] = dices[0]
        else:
            for i in range(self.n_time_courses):
                if self.names is not None:
                    key = "dice_" + self.names[i]
                else:
                    key = "dice" + str(i)
                logs[key] = dices[i]


class GumbelSoftmaxAnnealingCallback(tf.keras.callbacks.Callback):
    """Callback to anneal the temperature of a Gumbel-Softmax distribution.

    Parameters
    ----------
    curve : str
        Shape of the annealing curve.
        Can be either :code:`'linear'` or :code:`'exp'`.
    layer_name : str
        Name of the Gumbel-Softmax layer.
    n_epochs : int
        Total number of epochs.
    start_temperature : float, optional
        Starting temperature for the annealing.
    end_temperature : float, optional
        Ending temperature for the annealing.
    slope : float
        Slope of the curve. Only used when :code:`curve='exp'`.
    """

    def __init__(
        self,
        curve,
        layer_name,
        n_epochs,
        start_temperature=1.0,
        end_temperature=0.01,
        slope=0.014,
    ):
        self.curve = curve
        self.layer_name = layer_name
        self.n_epochs = n_epochs
        self.start_temperature = start_temperature
        self.end_temperature = end_temperature
        self.slope = slope

        # Precompute temperatures for linear decay
        if self.curve == "linear":
            self.temperatures = np.linspace(
                start_temperature, end_temperature, n_epochs
            )

    def set_model(self, model):
        # Cache the Gumbel-Softmax layer when the model is set
        super().set_model(model)
        self.gumbel_softmax_layer = model.get_layer(self.layer_name)

    def on_epoch_begin(self, epoch, logs=None):
        if self.curve == "linear":
            temperature = self.temperatures[epoch]
        if self.curve == "exp":
            temperature = max(
                self.end_temperature,
                self.start_temperature * np.exp(-self.slope * epoch),
            )

        self.gumbel_softmax_layer.temperature.assign(temperature)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["temperature"] = float(self.gumbel_softmax_layer.temperature.numpy())


class KLAnnealingCallback(callbacks.Callback):
    """Callback to update the KL annealing factor during training.

    This callback assumes there is a keras layer named :code:`'kl_loss'`
    in the model.

    Parameters
    ----------
    curve : str
        Shape of the annealing curve. Either :code:`'linear'` or :code:`'tanh'`.
    annealing_sharpness : float
        Parameter to control the shape of the annealing curve.
    n_annealing_epochs : int
        Number of epochs to apply annealing.
    n_cycles : int, optional
        Number of times to perform KL annealing with :code:`n_annealing_epochs`.
    """

    def __init__(
        self,
        curve,
        annealing_sharpness,
        n_annealing_epochs,
        n_cycles=1,
    ):
        if curve not in ["linear", "tanh"]:
            raise NotImplementedError(curve)

        super().__init__()
        self.curve = curve
        self.annealing_sharpness = annealing_sharpness
        self.n_annealing_epochs = n_annealing_epochs
        self.n_cycles = n_cycles
        self.n_epochs_one_cycle = n_annealing_epochs // n_cycles

    def on_epoch_end(self, epoch, logs=None):
        """Action to perform at the end of an epoch.

        Parameters
        ---------
        epochs : int
            Integer, index of epoch.
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """

        # Calculate new value
        epoch += 1  # epoch goes from 0 to n_epochs - 1, so we add 1
        if epoch < self.n_annealing_epochs:
            epoch = epoch % self.n_epochs_one_cycle
            if self.curve == "tanh":
                new_value = (
                    0.5
                    * tanh(
                        self.annealing_sharpness
                        * (epoch - 0.5 * self.n_epochs_one_cycle)
                        / self.n_epochs_one_cycle
                    )
                    + 0.5
                )
            elif self.curve == "linear":
                new_value = epoch / self.n_epochs_one_cycle
        else:
            new_value = 1.0

        # Update the annealing factor in the layer that calculates the KL loss
        kl_loss_layer = self.model.get_layer("kl_loss")
        kl_loss_layer.annealing_factor.assign(new_value)

        # Annealing factor for gamma sampling
        if "means_dev_mag" in self.model.layers:
            means_dev_mag_layer = self.model.get_layer("means_dev_mag")
            means_dev_mag_layer.annealing_factor.assign(new_value)

        if "covs_dev_mag" in self.model.layers:
            covs_dev_mag_layer = self.model.get_layer("covs_dev_mag")
            covs_dev_mag_layer.annealing_factor.assign(new_value)

        logs["kl_factor"] = new_value


class EMADecayCallback(callbacks.Callback):
    """Callback to update the decay rate in an Exponential Moving Average optimizer.

    :code:`decay = (100 * epoch / n_epochs + 1 + delay) ** -forget`

    Parameters
    ----------
    delay : float
    forget : float
    """

    def __init__(self, delay, forget, n_epochs):
        super().__init__()
        self.delay = delay
        self.forget = forget
        self.n_epochs = n_epochs

    def on_epoch_end(self, epoch, logs=None):
        """Action to perform at the end of an epoch.

        Parameters
        ---------
        epochs : int
            Integer, index of epoch.
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """

        # Calculate new value
        new_value = (100 * epoch / self.n_epochs + 1 + self.delay) ** -self.forget

        # Print new value during training
        logs["rho"] = new_value

        # Update the decay parameter in the optimizer
        # Here we are assuming a MarkovStateModelOptimizer is being used
        ema_optimizer = self.model.optimizer.ema_optimizer
        ema_optimizer.decay.assign(new_value)


class SaveBestCallback(callbacks.ModelCheckpoint):
    """Callback to save the best model.

    The best model is determined as the model with the lowest loss.

    Parameters
    ----------
    save_best_after : int
        Epoch number after which to save the best model.
    """

    def __init__(self, save_best_after, *args, **kwargs):
        self.save_best_after = save_best_after

        kwargs.update(
            dict(
                save_weights_only=True,
                monitor="loss",
                mode="min",
                save_best_only=True,
            )
        )

        super().__init__(*args, **kwargs)

    def on_epoch_end(self, epoch, logs=None):
        """Action to perform at the end of an epoch.

        Parameters
        ---------
        epochs : int
            Integer, index of epoch.
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """
        self.epochs_since_last_save += 1
        if epoch >= self.save_best_after:
            if self.save_freq == "epoch":
                self._save_model(epoch=epoch, logs=logs, batch=None)

    def on_train_end(self, logs=None):
        """Action to perform at the end of training.

        Parameters
        ----------
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """
        self.model.load_weights(self.filepath)


class CheckpointCallback(callbacks.Callback):
    """Callback to create checkpoints during training.

    Parameters
    ----------
    save_freq : int
        Frequency (in epochs) at which to save the model.
    """

    def __init__(self, save_freq, checkpoint_dir):
        super().__init__()
        self.save_freq = save_freq
        self.checkpoint = None
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_prefix = f"{checkpoint_dir}/ckpt"

    def on_epoch_end(self, epoch, logs=None):
        if self.checkpoint is None:
            self.checkpoint = tf.train.Checkpoint(
                model=self.model, optimizer=self.model.optimizer
            )
        if (epoch + 1) % self.save_freq == 0:
            self.checkpoint.save(file_prefix=self.checkpoint_prefix)


class TensorBoardCallback(callbacks.TensorBoard):
    """Callback to log training information to TensorBoard.

    This callback extends `tf.keras.callbacks.TensorBoard` by also logging the initial weights.

    Parameters
    ----------
    log_dir : str, optional
        Path to a directory where the log files will be written.
        Defaults to None, in which case the logs will be written to a current directory.
    log_initial : bool, optional
        Whether to log the initial weights or not. Defaults to True.
    step_offset : int, optional
        Offset to add to the epoch number when logging gradients. Defaults to 0.
    kwargs : dict
        Additional arguments to pass to the :code:`tf.keras.callbacks.TensorBoard` callback.
    """

    def __init__(self, log_dir=None, log_initial=True, step_offset=0, **kwargs):
        # Create log directory if it does not exist
        self._log_dir = log_dir
        self._make_log_dir()

        # Get arguments
        self.log_initial = log_initial  # enable or disable initial weight logging
        self.initial_weights_logged = False  # log status
        self.step_offset = step_offset  # offset to add to the epoch number

        super().__init__(log_dir=self._log_dir, **kwargs)

    def _make_log_dir(self):
        if self._log_dir is None:
            self._log_dir = os.path.join(os.getcwd(), "logs")
        os.makedirs(self._log_dir, exist_ok=True)

    def on_train_begin(self, logs=None):
        # Call the parent method first
        super().on_train_begin(logs)

        # Log the initial weights once
        if self.log_initial and not self.initial_weights_logged:
            # Create a subdirectory for the initial weights
            init_log_dir = os.path.join(self._log_dir, "initial_weights")
            os.makedirs(init_log_dir, exist_ok=True)
            writer = tf.summary.create_file_writer(init_log_dir)

            # Log the initial weights
            with writer.as_default():
                for weight in self.model.weights:
                    tf.summary.histogram(weight.name, weight, step=0)
                writer.flush()  # ensure all buffered data are written to disk

            self.initial_weights_logged = True
            print(
                "Initial weights logged. You can launch TensorBoard to view the histograms."
            )

    def on_epoch_end(self, epoch, logs=None):
        # Compute a continuous global step by adding an offset
        global_epoch = epoch + self.step_offset

        self._log_epoch_metrics(global_epoch, logs)

        if self.histogram_freq and global_epoch % self.histogram_freq == 0:
            self._log_weights(global_epoch)

        if self.embeddings_freq and global_epoch % self.embeddings_freq == 0:
            self._log_embeddings(global_epoch)


class GradientMonitoringCallback(tf.keras.callbacks.Callback):
    """Callback for logging gradients during the model training.

    Parameters
    ----------
    sample_dataset : tf.data.Dataset
        A dataset containing a representative batch of data used to compute gradients.
    loss_indices : int or list of int
        Indices of the losses in the model output.
    log_dir : str, optional
        Path to a directory where gradient logs will be saved.
        Defaults to None, in which case the logs will be written to the current directory.
    log_as_dense : bool, optional
        Whether to log gradients as dense tensors or not. Defaults to True.
        If False, only non-zero gradients will be logged (if the gradient is sparse).
    step_offset : int, optional
        Offset to add to the epoch number when logging gradients. Defaults to 0.
    print_stats : bool, optional
        Wheter to print the summary statistics (mean, std, min, max, L2 norm) for each variable.
        Defaults to False.
    """

    def __init__(
        self,
        sample_dataset,
        loss_indices,
        log_dir=None,
        log_as_dense=True,
        step_offset=0,
        print_stats=False,
    ):
        super().__init__()
        self.sample_dataset = sample_dataset
        self.loss_indices = loss_indices
        self.log_as_dense = log_as_dense
        self.step_offset = step_offset
        self.print_stats = print_stats

        # Validate inputs
        if isinstance(loss_indices, int):
            self.loss_indices = [loss_indices]

        # Prepare a log directory
        if log_dir is None:
            log_dir = os.path.join(os.getcwd(), "logs/gradients")
        os.makedirs(log_dir, exist_ok=True)
        self.writer = tf.summary.create_file_writer(log_dir)

    @tf.function
    def compute_gradients(self, inputs):
        """Compute gradients for a given input batch.
        If there is more than one loss, losses are summed before computing gradients.

        Parameters
        ----------
        inputs : tf.Tensor
            Input batch.
        """
        with tf.GradientTape() as tape:
            outputs = self.model(inputs, training=True)
            if len(self.loss_indices) > 1:
                loss = tf.add_n([outputs[idx] for idx in self.loss_indices])
            else:
                loss = outputs[self.loss_indices[0]]
        return tape.gradient(loss, self.model.trainable_variables)

    def _convert_grad_to_dense(self, gradient):
        """Convert a gradient to a dense tensor if necessary.

        Parameters
        ----------
        gradient : tf.Tensor, tf.IndexedSlices, tf.SparseTensor
            Gradient to convert to a dense tensor.

        Returns
        -------
        converted_gradient : tf.Tensor
            Dense tensor representation of the gradient.
        sparse_flag : bool
            Flag indicating whether the gradient was originally sparse.
        """
        if isinstance(gradient, tf.IndexedSlices):
            return tf.convert_to_tensor(gradient), True
        elif isinstance(gradient, tf.SparseTensor):
            return tf.sparse.to_dense(gradient), True
        return gradient, False

    def on_epoch_end(self, epoch, logs=None):
        """Action to perform at the end of an epoch.

        Parameters
        ---------
        epochs : int
            Index of epoch.
        logs : dict, optional
            Results for this training epoch, and for the validation epoch if
            validation is performed.
        """
        # Define logging step
        step = epoch + self.step_offset

        # Initialize accumulators for each trainable variable
        accumulated_gradients = [None] * len(self.model.trainable_variables)
        sparse_flags = [True] * len(self.model.trainable_variables)
        batch_count = 0

        # Compute the loss and gradients on the sample dataset
        for batch in self.sample_dataset:
            inputs = batch["data"]
            gradients = self.compute_gradients(inputs)

            # Accumulate gradients
            for i, grad in enumerate(gradients):
                # Always convert to dense for correct accumulation
                grad, sparse_flag = self._convert_grad_to_dense(grad)
                if grad is not None:
                    if accumulated_gradients[i] is None:
                        accumulated_gradients[i] = grad
                    else:
                        accumulated_gradients[i] += grad
                # If any batch gives a dense gradient, mark the overall flag as False
                if sparse_flags[i] is True:
                    sparse_flags[i] = sparse_flag
                    # ensure that non-zero values are removed only if all gradients are sparse
            batch_count += 1

        # Average gradients over the batches
        if batch_count > 0:
            averaged_gradients = [
                grad / batch_count if grad is not None else None
                for grad in accumulated_gradients
            ]
        else:
            averaged_gradients = accumulated_gradients

        # Group gradients by layer
        layer_gradients = {}
        for i, (grad, var) in enumerate(
            zip(averaged_gradients, self.model.trainable_variables)
        ):
            var_name = var.name
            layer_name = var_name.split("/")[0]  # get the first part as the layer name.
            if grad is not None:
                layer_gradients.setdefault(layer_name, []).append(
                    (grad, var, sparse_flags[i])
                )

        # Log and print gradient summary statistics for each layer
        with self.writer.as_default():
            for layer, grad_var_pairs in layer_gradients.items():
                if self.print_stats:
                    print(f"\nLayer: {layer}")
                for grad, var, flag in grad_var_pairs:
                    if grad is not None:
                        if not self.log_as_dense and flag:
                            # Log only non-zero entries, given that the gradient is sparse
                            logged_grad = tf.boolean_mask(grad, tf.not_equal(grad, 0))
                        else:
                            # Log the full dense gradient
                            logged_grad = grad

                        # Compute summary statistics
                        grad_mean = tf.reduce_mean(logged_grad)
                        grad_std = tf.math.reduce_std(logged_grad)
                        grad_min = tf.reduce_min(logged_grad)
                        grad_max = tf.reduce_max(logged_grad)
                        grad_norm = tf.norm(logged_grad)

                        # Compute statistics for non-zero entries
                        nonzero_mask = tf.not_equal(grad, 0)
                        nonzero_vals = tf.boolean_mask(grad, nonzero_mask)
                        nonzero_ratio = tf.cast(
                            tf.size(nonzero_vals), tf.float32
                        ) / tf.cast(tf.size(grad), tf.float32)

                        # Print summary statistics
                        if self.print_stats:
                            print(f"  {var.name}:")
                            print(
                                f"    Mean: {grad_mean.numpy():.5f}, Std: {grad_std.numpy():.5f}"
                            )
                            print(
                                f"    Min: {grad_min.numpy():.5f}, Max: {grad_max.numpy():.5f}"
                            )
                            print(f"    L2 Norm: {grad_norm.numpy():.5f}")
                            if flag:
                                print(
                                    f"    Non-zero ratio: {nonzero_ratio.numpy():.5f}"
                                )

                        # Log gradient histogram and scalar summaries
                        if not self.log_as_dense and flag:
                            tf.summary.histogram(
                                f"gradients/{var.name}_nonzero", logged_grad, step=step
                            )
                        else:
                            tf.summary.histogram(
                                f"gradients/{var.name}", grad, step=step
                            )
                        tf.summary.scalar(
                            f"gradients/{var.name}_mean", grad_mean, step=step
                        )
                        tf.summary.scalar(
                            f"gradients/{var.name}_std", grad_std, step=step
                        )
                        tf.summary.scalar(
                            f"gradients/{var.name}_min", grad_min, step=step
                        )
                        tf.summary.scalar(
                            f"gradients/{var.name}_max", grad_max, step=step
                        )
                        tf.summary.scalar(
                            f"gradients/{var.name}_norm", grad_norm, step=step
                        )
                        tf.summary.scalar(
                            f"gradients/{var.name}_nonzero_ratio",
                            nonzero_ratio,
                            step=step,
                        )
                    else:
                        if self.print_stats:
                            print(f"  {var.name}: Gradient is None.")
            self.writer.flush()



"""Multivariate autoregressive (MAR) observation model."""

import numpy as np


class MAR:
    r"""Class that generates data from a multivariate autoregressive (MAR) model.

    A :math:`p`-order MAR model can be written as

    .. math::
        x_t = A_1 x_{t-1} + ... + A_p x_{t-p} + \epsilon_t

    where :math:`\epsilon_t \sim N(0, \Sigma)`. The MAR model is therefore
    parameterized by the MAR coefficients (:math:`A`) and covariance
    (:math:`\Sigma`).

    Parameters
    ----------
    coeffs : np.ndarray
        Array of MAR coefficients, :math:`A`. Shape must be
        (n_states, n_lags, n_channels, n_channels).
    covs : np.ndarray
        Covariance of the error :math:`\epsilon_t`. Shape must be
        (n_states, n_channels) or (n_states, n_channels, n_channels).

    Note
    ----
    This model is also known as VAR or MVAR.
    """

    def __init__(self, coeffs, covs):
        # Validation
        if coeffs.ndim != 4:
            raise ValueError(
                "coeffs must be a (n_states, n_lags, n_channels, n_channels) array."
            )

        if covs.ndim == 2:
            covs = np.array([np.diag(c) for c in covs])

        if coeffs.shape[0] != covs.shape[0]:
            raise ValueError("Different number of states in coeffs and covs passed.")

        if coeffs.shape[-1] != covs.shape[-1]:
            raise ValueError("Different number of channels in coeffs and covs passed.")

        # Model parameters
        self.coeffs = coeffs
        self.covs = covs
        self.order = coeffs.shape[1]

        # Number of states and channels
        self.n_states = coeffs.shape[0]
        self.n_channels = coeffs.shape[2]

    def simulate_data(self, state_time_course):
        """Simulate time series data.

        We simulate MAR data based on the hidden state at each time point.

        Parameters
        ----------
        state_time_course : np.ndarray
            State time course. Shape must be (n_samples, n_states).
            States must be mutually exclusive.

        Returns
        -------
        data : np.ndarray
            Simulated data. Shape is (n_samples, n_channels).
        """
        n_samples = state_time_course.shape[0]
        data = np.empty([n_samples, self.n_channels])

        # Generate the noise term first
        for i in range(self.n_states):
            time_points_active = state_time_course[:, i] == 1
            n_time_points_active = np.count_nonzero(time_points_active)
            data[time_points_active] = np.random.multivariate_normal(
                np.zeros(self.n_channels),
                self.covs[i],
                size=n_time_points_active,
            )

        # Generate the MAR process
        for t in range(n_samples):
            state = state_time_course[t].argmax()
            for lag in range(min(t, self.order)):
                data[t] += np.dot(self.coeffs[state, lag], data[t - lag - 1])

        return data.astype(np.float32)



"""Files included within osl-dynamics.

The available files are listed in each submodule.
"""

from osl_dynamics.files import mask, parcellation, scanner, scene
from osl_dynamics.files.functions import *



"""Helpful functions related to observation models."""

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability import bijectors as tfb

import osl_dynamics.data.tf as dtf
from osl_dynamics.inference import regularizers
from osl_dynamics.inference.initializers import (
    WeightInitializer,
    RandomWeightInitializer,
)


def get_observation_model_parameter(model, layer_name):
    """Get the parameter of an observation model layer.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    layer_name : str
        Name of the layer of the observation model parameter.

    Returns
    -------
    obs_parameter : np.ndarray
        The observation model parameter.
    """
    available_layers = [
        "means",
        "covs",
        "stds",
        "corrs",
        "group_means",
        "group_covs",
        "log_rates",
    ]
    if layer_name not in available_layers:
        raise ValueError(
            f"Layer name {layer_name} not in available layers {available_layers}."
        )
    obs_layer = model.get_layer(layer_name)
    obs_parameter = obs_layer(tf.constant(1))
    return obs_parameter.numpy()


def set_observation_model_parameter(
    model,
    obs_parameter,
    layer_name,
    update_initializer=True,
    diagonal_covariances=False,
):
    """Set the value of an observation model parameter.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    obs_parameter : np.ndarray
        The value of the observation model parameter to set.
    layer_name : str
        Layer name of the observation model parameter.
    update_initializer : bool, optional
        Whether to update the initializer of the layer.
    diagonal_covariances : bool, optional
        Whether the covariances are diagonal.
        Ignored if :code:`layer_name` is not :code:`"covs"`.
    """
    available_layers = [
        "means",
        "covs",
        "stds",
        "corrs",
        "group_means",
        "group_covs",
        "log_rates",
    ]
    if layer_name not in available_layers:
        raise ValueError(
            f"Layer name {layer_name} not in available layers {available_layers}."
        )

    obs_parameter = obs_parameter.astype(np.float32)

    if layer_name == "stds" or (layer_name == "covs" and diagonal_covariances):
        if obs_parameter.ndim == 3:
            # Only keep the diagonal as a vector
            obs_parameter = np.diagonal(obs_parameter, axis1=1, axis2=2)

    obs_layer = model.get_layer(layer_name)
    learnable_tensor_layer = obs_layer.layers[0]

    if layer_name not in ["means", "group_means", "log_rates"]:
        obs_parameter = obs_layer.bijector.inverse(obs_parameter)

    learnable_tensor_layer.tensor.assign(obs_parameter)

    if update_initializer:
        learnable_tensor_layer.tensor_initializer = WeightInitializer(obs_parameter)


def set_dev_parameters_initializer(
    model, training_dataset, learn_means, learn_covariances
):
    """Set the deviance parameters initializer based on training data.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    training_dataset : tf.data.Dataset
        The training dataset.
    learn_means : bool
        Whether the mean is learnt.
    learn_covariances : bool
        Whether the covariances are learnt.
    """
    time_series = []
    for d in training_dataset:
        subject_data = []
        for batch in d:
            subject_data.append(np.concatenate(batch["data"]))
        time_series.append(np.concatenate(subject_data))

    n_channels = time_series[0].shape[1]
    if isinstance(time_series, np.ndarray):
        time_series = [time_series]
    if learn_means:
        static_means = np.array([np.mean(t, axis=0) for t in time_series])
        static_means_dev = np.abs(static_means - np.mean(static_means, axis=0))
        static_means_dev_mean = np.mean(static_means_dev, axis=1)
        static_means_dev_var = static_means_dev_mean / 5

        means_alpha = tfp.math.softplus_inverse(
            np.square(static_means_dev_mean) / static_means_dev_var
        )[..., None, None]
        means_beta = tfp.math.softplus_inverse(
            static_means_dev_mean / static_means_dev_var
        )[..., None, None]

        means_alpha_layer = model.get_layer("means_dev_mag_inf_alpha_input")
        means_beta_layer = model.get_layer("means_dev_mag_inf_beta_input")

        means_alpha_layer.tensor_initializer = RandomWeightInitializer(means_alpha, 0.1)
        means_beta_layer.tensor_initializer = RandomWeightInitializer(means_beta, 0.1)

    if learn_covariances:
        static_cov = np.array([np.cov(t, rowvar=False) for t in time_series])
        static_cov_chol = np.linalg.cholesky(static_cov)[
            :,
            np.tril_indices(n_channels)[0],
            np.tril_indices(n_channels)[1],
        ]
        static_cov_chol_dev = np.abs(static_cov_chol - np.mean(static_cov_chol, axis=0))
        static_cov_chol_dev_mean = np.mean(static_cov_chol_dev, axis=1)
        static_cov_chol_dev_var = static_cov_chol_dev_mean / 5

        covs_alpha = tfp.math.softplus_inverse(
            np.square(static_cov_chol_dev_mean) / static_cov_chol_dev_var
        )[..., None, None]
        covs_beta = tfp.math.softplus_inverse(
            static_cov_chol_dev_mean / static_cov_chol_dev_var
        )[..., None, None]

        covs_alpha_layer = model.get_layer("covs_dev_mag_inf_alpha_input")
        covs_beta_layer = model.get_layer("covs_dev_mag_inf_beta_input")

        covs_alpha_layer.tensor_initializer = RandomWeightInitializer(covs_alpha, 0.1)
        covs_beta_layer.tensor_initializer = RandomWeightInitializer(covs_beta, 0.1)


def set_embeddings_initializer(model, initial_embeddings):
    """Set the embeddings initializer.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    initial_embeddings : dict
        The initial_embeddings dictionary. {name: value}
    """

    # Helper function to set a single layer's initializer
    def _set_embeddings_initializer(layer_name, value):
        embedding_layer = model.get_layer(layer_name)
        embedding_layer.embedding_layer.embeddings_initializer = WeightInitializer(
            value
        )

    for k, v in initial_embeddings.items():
        _set_embeddings_initializer(f"{k}_embeddings", v)


def set_means_regularizer(model, range_, scale_factor, layer_name="means"):
    """Set the means regularizer based on training data.

    A multivariate normal prior is applied to the mean vectors with
    :code:`mu=0`, :code:`sigma=diag((range/2)**2)`.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    range_ : np.ndarray
        Range (max-min) of the training data for each channel.
        Shape is (n_channels,).
    scale_factor : float
        Scale factor for regularization.
    layer_name : str, optional
        Layer name of the means. Can be :code:`"means"` or
        :code:`"group_means"`.
    """
    n_channels = range_.shape[0]
    mu = np.zeros(n_channels, dtype=np.float32)
    sigma = np.diag((range_ / 2) ** 2)
    means_layer = model.get_layer(layer_name)
    learnable_tensor_layer = means_layer.layers[0]
    learnable_tensor_layer.regularizer = regularizers.MultivariateNormal(
        mu, sigma, scale_factor
    )


def set_covariances_regularizer(
    model,
    range_,
    epsilon,
    scale_factor,
    diagonal=False,
    layer_name="covs",
):
    """Set the covariances regularizer based on training data.

    If config.diagonal_covariances is True, a log-normal prior is applied to
    the diagonal of the covariance matrices with :code:`mu=0`,
    :code:`sigma=sqrt(log(2*range))`. Otherwise, an inverse Wishart prior is
    applied to the covariance matrices with :code:`nu=n_channels-1+0.1`,
    :code:`psi=diag(1/range)`.x

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    range_ : np.ndarray
        Range (max-min) of the training data for each channel.
        Shape is (n_channels,).
    epsilon : float
        Error added to the covariance matrices.
    scale_factor : float
        Scale factor for regularization.
    diagonal : bool, optional
        Whether the covariances are diagonal.
    layer_name : str, optional
        Layer name of the covariances. Can be :code:`"covs"` or
        :code:`"group_covs"`.
    """
    n_channels = range_.shape[0]
    covs_layer = model.get_layer(layer_name)
    if diagonal:
        mu = np.zeros([n_channels], dtype=np.float32)
        sigma = np.sqrt(np.log(2 * range_))
        learnable_tensor_layer = covs_layer.layers[0]
        learnable_tensor_layer.regularizer = regularizers.LogNormal(
            mu, sigma, epsilon, scale_factor
        )
    else:
        nu = n_channels - 1 + 0.1
        psi = np.diag(range_)
        learnable_tensor_layer = covs_layer.layers[0]
        learnable_tensor_layer.regularizer = regularizers.InverseWishart(
            nu, psi, epsilon, scale_factor
        )


def set_stds_regularizer(model, range_, epsilon, scale_factor, layer_name="stds"):
    """Set the standard deviations regularizer based on training data.

    A log-normal prior is applied to the standard deviations with :code:`mu=0`,
    :code:`sigma=sqrt(log(2*range))`.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    range_ : np.ndarray
        Range (max-min) of the training data for each channel.
        Shape is (n_channels,).
    epsilon : float
        Error added to the standard deviations.
    scale_factor : float
        Scale factor for regularization.
    layer_name : str, optional
        Layer name of the covariances.
    """
    n_channels = range_.shape[0]
    mu = np.zeros([n_channels], dtype=np.float32)
    sigma = np.sqrt(np.log(2 * range_))
    stds_layer = model.get_layer(layer_name)
    learnable_tensor_layer = stds_layer.layers[0]
    learnable_tensor_layer.regularizer = regularizers.LogNormal(
        mu, sigma, epsilon, scale_factor
    )


def set_corrs_regularizer(model, n_channels, epsilon, scale_factor, layer_name="corrs"):
    """Set the correlations regularizer based on training data.

    A marginal inverse Wishart prior is applied to the correlations
    with :code:`nu=n_channels-1+0.1`.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model.
    range_ : np.ndarray
        Range (max-min) of the training data for each channel.
        Shape is (n_channels,).
    epsilon : float
        Error added to the correlations.
    scale_factor : float
        Scale factor for regularization.
    layer_name : str, optional
        Layer name of the covariances.
    """
    nu = n_channels - 1 + 0.1
    corrs_layer = model.get_layer(layer_name)
    learnable_tensor_layer = corrs_layer.layers[0]
    learnable_tensor_layer.regularizer = regularizers.MarginalInverseWishart(
        nu, epsilon, n_channels, scale_factor
    )


def get_embedding_weights(model, session_labels):
    """Get the weights of the embedding layers.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    embedding_weights : dict
        The weights of the embedding layers.
    """
    embedding_weights = dict()
    for session_label in session_labels:
        label_name = session_label.name
        label_type = session_label.label_type
        embeddings_layer = model.get_layer(f"{label_name}_embeddings")
        if label_type == "categorical":
            embedding_weights[label_name] = embeddings_layer.embeddings.numpy()
        else:
            embedding_weights[label_name] = [
                embeddings_layer.kernel.numpy(),
                embeddings_layer.bias.numpy(),
            ]

    return embedding_weights


def get_session_embeddings(model, session_labels):
    """Get the embeddings for each session.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    embeddings : dict
        The embeddings for each session label.
    """
    embeddings = dict()
    for session_label in session_labels:
        label_name = session_label.name
        label_values = session_label.values
        embeddings_layer = model.get_layer(f"{label_name}_embeddings")
        embeddings[label_name] = embeddings_layer(label_values)

    return embeddings


def get_summed_embeddings(model, session_labels):
    """Get the summed embeddings for each session.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    summed_embeddings : np.ndarray
        The summed embeddings. Shape is (n_sessions, embeddings_dim).
    """
    embeddings = get_session_embeddings(model, session_labels)
    summed_embeddings = 0
    for _, embedding in embeddings.items():
        summed_embeddings += embedding
    return summed_embeddings.numpy()


def get_means_spatial_embeddings(model):
    """Get the means spatial embeddings.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.

    Returns
    -------
    means_spatial_embeddings : np.ndarray
        The means spatial embeddings. Shape is (n_states, spatial_embeddings_dim).
    """
    group_means = get_observation_model_parameter(model, "group_means")
    means_spatial_embeddings_layer = model.get_layer("means_spatial_embeddings")
    means_spatial_embeddings = means_spatial_embeddings_layer(group_means)
    return means_spatial_embeddings.numpy()


def get_covs_spatial_embeddings(model):
    """Get the covariances spatial embeddings.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.

    Returns
    -------
    covs_spatial_embeddings : np.ndarray
        The covariances spatial embeddings.
        Shape is (n_states, spatial_embeddings_dim).
    """
    cholesky_bijector = tfb.Chain([tfb.CholeskyOuterProduct(), tfb.FillScaleTriL()])
    group_covs = get_observation_model_parameter(model, "group_covs")
    covs_spatial_embeddings_layer = model.get_layer("covs_spatial_embeddings")
    covs_spatial_embeddings = covs_spatial_embeddings_layer(
        cholesky_bijector.inverse(group_covs)
    )
    return covs_spatial_embeddings.numpy()


def get_spatial_embeddings(model, param):
    """Wrapper for getting the spatial embeddings for the means and covariances."""
    if param == "means":
        return get_means_spatial_embeddings(model)
    elif param == "covs":
        return get_covs_spatial_embeddings(model)
    else:
        raise ValueError("param must be either 'means' or 'covs'")


def get_concatenated_embeddings(model, param, session_labels):
    """Get the concatenated embeddings.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    param : str
        The param to use. Either :code:`"means"` or :code:`"covs"`.
    embeddings : np.ndarray, optional
        Input embeddings. If :code:`None`, they are retrieved from
        the model. Shape is (n_sessions, embeddings_dim).

    Returns
    -------
    concat_embeddings : np.ndarray
        The concatenated embeddings. Shape is (n_sessions, n_states,
        embeddings_dim + spatial_embeddings_dim).
    """
    embeddings = get_summed_embeddings(model, session_labels)
    if param == "means":
        spatial_embeddings = get_means_spatial_embeddings(model)
        concat_embeddings_layer = model.get_layer("means_concat_embeddings")
    elif param == "covs":
        spatial_embeddings = get_covs_spatial_embeddings(model)
        concat_embeddings_layer = model.get_layer("covs_concat_embeddings")
    else:
        raise ValueError("param must be either 'means' or 'covs'")
    concat_embeddings = concat_embeddings_layer([embeddings, spatial_embeddings])
    return concat_embeddings.numpy()


def get_means_dev_mag_parameters(model):
    """Get the means deviation magnitude parameters.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.

    Returns
    -------
    means_dev_mag_inf_alpha : np.ndarray
        The means deviation magnitude alpha parameters.
        Shape is (n_sessions, n_states, 1).
    means_dev_mag_inf_beta : np.ndarray
        The means deviation magnitude beta parameters.
        Shape is (n_sessions, n_states, 1).
    """
    means_dev_mag_inf_alpha_input_layer = model.get_layer(
        "means_dev_mag_inf_alpha_input"
    )
    means_dev_mag_inf_alpha_layer = model.get_layer("means_dev_mag_inf_alpha")
    means_dev_mag_inf_beta_input_layer = model.get_layer("means_dev_mag_inf_beta_input")
    means_dev_mag_inf_beta_layer = model.get_layer("means_dev_mag_inf_beta")

    means_dev_mag_inf_alpha_input = means_dev_mag_inf_alpha_input_layer(tf.constant(1))
    means_dev_mag_inf_alpha = means_dev_mag_inf_alpha_layer(
        means_dev_mag_inf_alpha_input
    )

    means_dev_mag_inf_beta_input = means_dev_mag_inf_beta_input_layer(tf.constant(1))
    means_dev_mag_inf_beta = means_dev_mag_inf_beta_layer(means_dev_mag_inf_beta_input)
    return means_dev_mag_inf_alpha.numpy(), means_dev_mag_inf_beta.numpy()


def get_covs_dev_mag_parameters(model):
    """Get the covariances deviation magnitude parameters.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.

    Returns
    -------
    covs_dev_mag_inf_alpha : np.ndarray
        The covariances deviation magnitude alpha parameters.
        Shape is (n_sessions, n_states, 1).
    covs_dev_mag_inf_beta : np.ndarray
        The covariances deviation magnitude beta parameters.
        Shape is (n_sessions, n_states, 1).
    """
    covs_dev_mag_inf_alpha_input_layer = model.get_layer("covs_dev_mag_inf_alpha_input")
    covs_dev_mag_inf_alpha_layer = model.get_layer("covs_dev_mag_inf_alpha")
    covs_dev_mag_inf_beta_input_layer = model.get_layer("covs_dev_mag_inf_beta_input")
    covs_dev_mag_inf_beta_layer = model.get_layer("covs_dev_mag_inf_beta")

    covs_dev_mag_inf_alpha_input = covs_dev_mag_inf_alpha_input_layer(tf.constant(1))
    covs_dev_mag_inf_alpha = covs_dev_mag_inf_alpha_layer(covs_dev_mag_inf_alpha_input)

    covs_dev_mag_inf_beta_input = covs_dev_mag_inf_beta_input_layer(tf.constant(1))
    covs_dev_mag_inf_beta = covs_dev_mag_inf_beta_layer(covs_dev_mag_inf_beta_input)
    return covs_dev_mag_inf_alpha.numpy(), covs_dev_mag_inf_beta.numpy()


def get_dev_mag_parameters(model, param):
    """Wrapper for getting the deviance magnitude parameters for the means
    and covariances."""
    if param == "means":
        return get_means_dev_mag_parameters(model)
    elif param == "covs":
        return get_covs_dev_mag_parameters(model)
    else:
        raise ValueError("param must be either 'means' or 'covs'")


def get_dev_mag(model, param):
    """Getting the deviance magnitude.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    param : str
        The param. Must be either :code:`'means'` or :code:`'covs'`.

    Returns
    -------
    dev_mag : np.ndarray
        The deviance magnitude. Shape is (n_sessions, n_states, 1).
    """
    if param == "means":
        alpha, beta = get_means_dev_mag_parameters(model)
        dev_mag_layer = model.get_layer("means_dev_mag")
    elif param == "covs":
        alpha, beta = get_covs_dev_mag_parameters(model)
        dev_mag_layer = model.get_layer("covs_dev_mag")
    else:
        raise ValueError("param must be either 'means' or 'covs'")

    n_sessions = alpha.shape[0]
    dev_mag = dev_mag_layer([alpha, beta, np.arange(n_sessions)[..., None]])
    return dev_mag.numpy()


def get_dev_map(model, param, session_labels):
    """Get the deviance map.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    param : str
        The param to use. Either :code:`"means"` or :code:`"covs"`.
    embeddings : np.ndarray, optional
        Input embeddings. If :code:`None`, they are retrieved from
        the model. Shape is (n_sessions, embeddings_dim).

    Returns
    -------
    dev_map : np.ndarray
        The deviance map.
        If :code:`param="means"`, shape is (n_sessions, n_states, n_channels).
        If :code:`param="covs"`, shape is (n_sessions, n_states,
        n_channels * (n_channels + 1) // 2).
    """
    concat_embeddings = get_concatenated_embeddings(model, param, session_labels)
    if param == "means":
        dev_decoder_layer = model.get_layer("means_dev_decoder")
        dev_map_layer = model.get_layer("means_dev_map")
        norm_dev_map_layer = model.get_layer("norm_means_dev_map")
    elif param == "covs":
        dev_decoder_layer = model.get_layer("covs_dev_decoder")
        dev_map_layer = model.get_layer("covs_dev_map")
        norm_dev_map_layer = model.get_layer("norm_covs_dev_map")
    else:
        raise ValueError("param must be either 'means' or 'covs'")
    dev_decoder = dev_decoder_layer(concat_embeddings)
    dev_map = dev_map_layer(dev_decoder)
    norm_dev_map = norm_dev_map_layer(dev_map)
    return norm_dev_map.numpy()


def get_session_dev(
    model,
    learn_means,
    learn_covariances,
    session_labels,
):
    """Get the session deviation.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    learn_means : bool
        Whether the mean is learnt.
    learn_covariances : bool
        Whether the covariances are learnt.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    means_dev : np.ndarray
        The means deviation. Shape is (n_sessions, n_states, n_channels).
    covs_dev : np.ndarray
        The covariances deviation.
        Shape is (n_sessions, n_states, n_channels * (n_channels + 1) // 2).
    """
    means_dev_layer = model.get_layer("means_dev")
    covs_dev_layer = model.get_layer("covs_dev")

    if learn_means:
        means_dev_mag = get_dev_mag(model, "means")
        means_dev_map = get_dev_map(model, "means", session_labels)
        means_dev = means_dev_layer([means_dev_mag, means_dev_map])
    else:
        means_dev = means_dev_layer(tf.constant(1))

    if learn_covariances:
        covs_dev_mag = get_dev_mag(model, "covs")
        covs_dev_map = get_dev_map(model, "covs", session_labels)
        covs_dev = covs_dev_layer([covs_dev_mag, covs_dev_map])
    else:
        covs_dev = covs_dev_layer(tf.constant(1))

    return means_dev.numpy(), covs_dev.numpy()


def get_session_means_covariances(
    model,
    learn_means,
    learn_covariances,
    session_labels,
):
    """Get the session means and covariances.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    learn_means : bool
        Whether the mean is learnt.
    learn_covariances : bool
        Whether the covariances are learnt.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    mu : np.ndarray
        The session means. Shape is (n_sessions, n_states, n_channels).
    D : np.ndarray
        The session covariances.
        Shape is (n_sessions, n_states, n_channels, n_channels).
    """
    group_means = get_observation_model_parameter(model, "group_means")
    group_covs = get_observation_model_parameter(model, "group_covs")
    means_dev, covs_dev = get_session_dev(
        model, learn_means, learn_covariances, session_labels
    )

    session_means_layer = model.get_layer("session_means")
    session_covs_layer = model.get_layer("session_covs")

    mu = session_means_layer([group_means, means_dev])
    D = session_covs_layer([group_covs, covs_dev])
    return mu.numpy(), D.numpy()


def generate_covariances(model, session_labels):
    """Generate covariances from the generative model.

    Parameters
    ----------
    model : osl_dynamics.models.*.Model.model
        The model. * must be :code:`hive` or :code:`dive`.
    session_labels : List[osl_dynamics.data.SessionLabel]
        List of session labels.

    Returns
    -------
    covs : np.ndarray
        The covariances. Shape is (n_sessions, n_states, n_channels, n_channels)
        or (n_states, n_channels, n_channels).
    """

    dev_map = get_dev_map(model, "covs", session_labels)
    concat_embeddings = get_concatenated_embeddings(model, "covs", session_labels)

    covs_dev_decoder_layer = model.get_layer("covs_dev_decoder")
    dev_mag_mod_layer = model.get_layer("covs_dev_mag_mod_beta")
    dev_mag_mod = tf.constant(1, dtype=tf.float32) / dev_mag_mod_layer(
        covs_dev_decoder_layer(concat_embeddings)
    )

    # Generate deviations
    dev_layer = model.get_layer("covs_dev")
    dev = dev_layer([dev_mag_mod, dev_map])

    # Generate covariances
    group_covs = get_observation_model_parameter(model, "group_covs")
    covs_layer = model.get_layer("session_covs")
    covs = np.squeeze(covs_layer([group_covs, dev]).numpy())

    return covs



"""Function related to TensorFlow datasets."""

import logging
import numpy as np

from osl_dynamics.utils import misc

_logger = logging.getLogger("osl-dynamics")


def get_n_sequences(arr, sequence_length, step_size=None):
    """Calculate the number of sequences an array will be split into.

    Parameters
    ----------
    arr : np.ndarray
        Time series data.
    sequence_length : int
        Length of sequences which the data will be segmented in to.
    step_size : int, optional
        The number of samples by which to move the sliding window between
        sequences.

    Returns
    -------
    n : int
        Number of sequences.
    """
    n_samples = arr.shape[0]

    # Number of non-overlapping sequences
    n_sequences = n_samples // sequence_length

    if step_size is not None:
        # Number of overlapping sequences
        n_sequences += (n_samples - step_size) // sequence_length

    return n_sequences


def concatenate_datasets(datasets):
    """Concatenates a list of TensorFlow datasets.

    Parameters
    ----------
    datasets : list
        List of TensorFlow datasets.

    Returns
    -------
    full_dataset : tf.data.Dataset
        Concatenated dataset.
    """
    full_dataset = datasets[0]
    for ds in datasets[1:]:
        full_dataset = full_dataset.concatenate(ds)
    return full_dataset


def create_dataset(data, sequence_length, step_size):
    """Creates a TensorFlow dataset of batched time series data.

    Parameters
    ----------
    data : dict
        Dictionary containing data to batch. Keys correspond to the input name
        for the model and the value is the data.
    sequence_length : int
        Sequence length to batch the data.
    step_size : int
        Number of samples to slide the sequence across the data.

    Returns
    -------
    dataset : tf.data.Dataset
        TensorFlow dataset.
    """
    from tensorflow.data import Dataset  # moved here to avoid slow imports

    # Generate a non-overlapping sequence dataset
    if step_size == sequence_length:
        dataset = Dataset.from_tensor_slices(data)
        dataset = dataset.batch(sequence_length, drop_remainder=True)

    # Create an overlapping multiple model input dataset
    else:

        def batch_windows(*windows):
            batched = [w.batch(sequence_length, drop_remainder=True) for w in windows]
            return Dataset.zip(tuple(batched))

        def tuple_to_dict(*d):
            names = list(data.keys())
            inputs = {}
            for i in range(len(data)):
                inputs[names[i]] = d[i]
            return inputs

        dataset = tuple([Dataset.from_tensor_slices(v) for v in data.values()])
        dataset = Dataset.zip(dataset)
        dataset = dataset.window(
            sequence_length,
            step_size,
            drop_remainder=True,
        )
        dataset = dataset.flat_map(batch_windows)
        dataset = dataset.map(tuple_to_dict)

    return dataset


def save_tfrecord(data, sequence_length, step_size, filepath):
    """Save dataset to a TFRecord file.

    Parameters
    ----------
    data : dict
        Dictionary containing data to batch. Keys correspond to the input name
        for the model and the value is the data.
    sequence_length : int
        Sequence length to batch the data.
    step_size : int
        Number of samples to slide the sequence across the data.
    filepath : str
        Path to save the TFRecord file.
    """
    import tensorflow as tf  # moved here to avoid slow imports
    from tensorflow.train import Feature, Features, Example, BytesList

    dataset = create_dataset(data, sequence_length, step_size)

    # Helper function to serialize a sequence to a tensorflow example
    # byte string
    def _make_example(sequence):
        # Note this function assumes all features are tf tensors
        # and can be converted to bytes
        features = Features(
            feature={
                k: Feature(
                    bytes_list=BytesList(
                        value=[tf.io.serialize_tensor(v).numpy()],
                    )
                )
                for k, v in sequence.items()
            }
        )
        return Example(features=features).SerializeToString()

    # Serialize each sequence and write to a TFRecord file
    with tf.io.TFRecordWriter(filepath) as writer:
        for sequence in dataset:
            writer.write(_make_example(sequence))


def load_tfrecord_dataset(
    tfrecord_dir,
    batch_size,
    shuffle=True,
    concatenate=True,
    drop_last_batch=False,
    repeat_count=1,
    buffer_size=4000,
    keep=None,
):
    """Load a TFRecord dataset.

    Parameters
    ----------
    tfrecord_dir : str
        Directory containing the TFRecord datasets.
    batch_size : int
        Number sequences in each mini-batch which is used to train the model.
    shuffle : bool, optional
        Should we shuffle sequences (within a batch) and batches.
    concatenate : bool, optional
        Should we concatenate the datasets for each array?
    drop_last_batch : bool, optional
        Should we drop the last batch if it is smaller than the batch size?
    repeat_count : int, optional
        Number of times to repeat the dataset. Default is 1.
    buffer_size : int, optional
        Buffer size for shuffling a TensorFlow Dataset. Smaller values will lead
        to less random shuffling but will be quicker. Default is 100000.
    keep : list of int, optional
        List of session indices to keep. If :code:`None`, then all sessions
        are kept.

    Returns
    -------
    dataset : tf.data.TFRecordDataset or tuple of tf.data.TFRecordDataset
        Dataset for training or evaluating the model along with the
        validation set :code:`validation_split` is present in the config.
    """
    import tensorflow as tf  # moved here to avoid slow imports

    tfrecord_config = misc.load(f"{tfrecord_dir}/tfrecord_config.pkl")
    identifier = tfrecord_config["identifier"]
    validation_split = tfrecord_config["validation_split"]
    n_sessions = tfrecord_config["n_sessions"]
    input_shapes = tfrecord_config["input_shapes"]

    keep = keep or list(range(n_sessions))

    # Helper functions
    def _parse_example(example):
        feature_description = {
            name: tf.io.FixedLenFeature([], tf.string) for name in input_shapes.keys()
        }
        parsed_example = tf.io.parse_single_example(
            example,
            feature_description,
        )
        return {
            name: tf.ensure_shape(
                tf.io.parse_tensor(tensor, tf.float32),
                input_shapes[name],
            )
            for name, tensor in parsed_example.items()
        }

    def _create_dataset(filenames, shuffle=shuffle, repeat_count=repeat_count):
        if concatenate:
            filenames = tf.data.Dataset.from_tensor_slices(filenames)

            if shuffle:
                # First shuffle the shards
                filenames = filenames.shuffle(len(filenames))

                # Create the TFRecord dataset
                full_dataset = filenames.interleave(
                    tf.data.TFRecordDataset, num_parallel_calls=tf.data.AUTOTUNE
                )

                # Parse the examples
                full_dataset = full_dataset.map(_parse_example)

                # Shuffle sequences
                full_dataset = full_dataset.shuffle(buffer_size)

                # Group into batches
                full_dataset = full_dataset.batch(
                    batch_size, drop_remainder=drop_last_batch
                )

                # Shuffle batches
                full_dataset = full_dataset.shuffle(buffer_size)

            else:
                # Create the TFRecord dataset
                full_dataset = filenames.interleave(
                    tf.data.TFRecordDataset, num_parallel_calls=tf.data.AUTOTUNE
                )

                # Parse the examples
                full_dataset = full_dataset.map(_parse_example)

                # Group into batches
                full_dataset = full_dataset.batch(
                    batch_size, drop_remainder=drop_last_batch
                )

            # Repeat the dataset
            full_dataset = full_dataset.repeat(repeat_count)

            return full_dataset.prefetch(tf.data.AUTOTUNE)

        # Otherwise create a dataset for each array separately
        else:
            full_datasets = []
            for filename in filenames:
                ds = tf.data.TFRecordDataset(filename)

                # Parse the examples
                ds = ds.map(_parse_example)

                if shuffle:
                    # Shuffle sequences
                    ds = ds.shuffle(buffer_size)

                # Group into batches
                ds = ds.batch(batch_size, drop_remainder=drop_last_batch)

                if shuffle:
                    # Shuffle batches
                    ds = ds.shuffle(buffer_size)

                # Repeat the dataset
                ds = ds.repeat(repeat_count)

                full_datasets.append(ds.prefetch(tf.data.AUTOTUNE))

            return full_datasets

    # Path to TFRecord files
    tfrecord_path = (
        f"{tfrecord_dir}"
        "/dataset-{val}_{array:0{v}d}-of-{n_session:0{v}d}"
        f"_{identifier}.tfrecord"
    )

    if validation_split is None:
        # Only create one dataset
        filenames = []
        for i in keep:
            filepath = tfrecord_path.format(
                array=i,
                val=0,
                n_session=n_sessions - 1,
                v=len(str(n_sessions - 1)),
            )
            filenames.append(filepath)

        return _create_dataset(filenames)

    else:
        # Create two datasets
        train_filenames = []
        val_filenames = []
        for i in keep:
            filepath = tfrecord_path.format(
                array=i,
                val=0,
                n_session=n_sessions - 1,
                v=len(str(n_sessions - 1)),
            )
            train_filenames.append(filepath)
            filepath = tfrecord_path.format(
                array=i,
                val=1,
                n_session=n_sessions - 1,
                v=len(str(n_sessions - 1)),
            )
            val_filenames.append(filepath)

        return _create_dataset(train_filenames), _create_dataset(
            val_filenames, shuffle=False, repeat_count=1
        )


def _validate_tf_dataset(dataset):
    """Check if the input is a valid TensorFlow dataset.

    Parameters
    ----------
    dataset : tf.data.Dataset or list
        TensorFlow dataset or list of datasets.

    Returns
    -------
    dataset : tf.data.Dataset
        TensorFlow dataset.
    """
    import tensorflow as tf  # avoid slow imports

    if isinstance(dataset, list):
        if len(dataset) == 1:
            dataset = dataset[0]
        else:
            dataset = concatenate_datasets(dataset)

    if not isinstance(dataset, tf.data.Dataset):
        raise ValueError("dataset must be a TensorFlow dataset or a list of datasets")

    return dataset


def get_range(dataset):
    """The range (max-min) of values contained in a batched Tensorflow dataset.

    Parameters
    ----------
    dataset : tf.data.Dataset
        TensorFlow dataset.

    Returns
    -------
    range_ : np.ndarray
        Range of each channel.
    """
    amax = []
    amin = []
    dataset = _validate_tf_dataset(dataset)
    for batch in dataset:
        if isinstance(batch, dict):
            batch = batch["data"]
        batch = batch.numpy()
        n_channels = batch.shape[-1]
        batch = batch.reshape(-1, n_channels)
        amin.append(np.amin(batch, axis=0))
        amax.append(np.amax(batch, axis=0))
    return np.amax(amax, axis=0) - np.amin(amin, axis=0)


def get_n_channels(dataset):
    """Get the number of channels in a batched TensorFlow dataset.

    Parameters
    ----------
    dataset : tf.data.Dataset
        TensorFlow dataset.

    Returns
    -------
    n_channels : int
        Number of channels.
    """
    dataset = _validate_tf_dataset(dataset)
    for batch in dataset:
        if isinstance(batch, dict):
            batch = batch["data"]
        batch = batch.numpy()
        return batch.shape[-1]


def get_n_batches(dataset):
    """Get number of batches in a TensorFlow dataset.

    Parameters
    ----------
    dataset : tf.data.Dataset
        TensorFlow dataset.

    Returns
    -------
    n_batches : int
        Number of batches.
    """
    import tensorflow as tf  # avoid slow imports

    dataset = _validate_tf_dataset(dataset)

    # Count number of batches
    cardinality = dataset.cardinality()
    if cardinality == tf.data.UNKNOWN_CARDINALITY:
        for i, _ in enumerate(dataset):
            pass
        return i + 1

    return cardinality.numpy()


def get_n_sequences_and_range(dataset):
    """Get number of sequences and range (max-min) of values.

    Parameters
    ----------
    dataset : tf.data.Dataset
        TensorFlow dataset.

    Returns
    -------
    n_sequences : int
        Number of batches.
    range_ : np.ndarray
        Range of each channel.
    """
    n_sequences = 0
    amax = []
    amin = []
    dataset = _validate_tf_dataset(dataset)
    for batch in dataset:
        if isinstance(batch, dict):
            batch = batch["data"]
        batch = batch.numpy()
        batch_size, _, n_channels = batch.shape
        batch = batch.reshape(-1, n_channels)
        amin.append(np.amin(batch, axis=0))
        amax.append(np.amax(batch, axis=0))
        n_sequences += batch_size
    range_ = np.amax(amax, axis=0) - np.amin(amin, axis=0)
    return n_sequences, range_



"""Functions to manipulate and analyse inferred mode/state data."""

import itertools
import logging

import numpy as np
from scipy import signal
from pqdm.threads import pqdm
from tqdm.auto import trange

from osl_dynamics import array_ops

_logger = logging.getLogger("osl-dynamics")


def autocorr_from_tde_cov(
    covs, n_embeddings, pca_components=None, sampling_frequency=None
):
    """Auto/cross-correlation function from the mode covariance matrices.

    Parameters
    ----------
    covs : np.ndarray
        Covariance matrix of time-delay embedded data. Shape must be
        (n_channels, n_channels) or (n_modes, n_channels, n_channels).
    n_embeddings : int
        Number of embeddings.
    pca_components : np.ndarray, optional
        PCA components used for dimensionality reduction. Only needs to be
        passed if PCA was performed on the time embedded data.
    sampling_frequency : float, optional
        Sampling_frequency in Hz.

    Returns
    -------
    tau : np.ndarray
        Time lags in samples if `sampling_frequency=None`, otherwise in seconds.
        Shape is (n_lags).
    acfs : np.ndarray
        Auto/cross-correlation functions. Shape is (n_channels, n_channels,
        n_lags) or (n_modes, n_channels, n_channels, n_lags).
    """
    # Validation
    error_message = (
        "covs must be of shape (n_channels, n_channels) or "
        + "(n_modes, n_channels, n_channels) or "
        + "(n_sessions, n_modes, n_channels, n_channels)."
    )
    covs = array_ops.validate(
        covs,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message=error_message,
    )

    if sampling_frequency is None:
        sampling_frequency = 1

    # Get covariance of time embedded data
    if pca_components is not None:
        te_covs = reverse_pca(covs, pca_components)
    else:
        te_covs = covs

    # Dimensions
    n_sessions = te_covs.shape[0]
    n_modes = te_covs.shape[1]
    n_parcels = te_covs.shape[-1] // n_embeddings
    n_lags = 2 * n_embeddings - 1

    # Take mean of elements from the time embedded covariances that
    # correspond to the auto/cross-correlation function
    blocks = te_covs.reshape(
        n_sessions,
        n_modes,
        n_parcels,
        n_embeddings,
        n_parcels,
        n_embeddings,
    )
    acfs = np.empty([n_sessions, n_modes, n_parcels, n_parcels, n_lags])
    for i in range(n_lags):
        acfs[:, :, :, :, i] = np.mean(
            np.diagonal(blocks, offset=i - n_embeddings + 1, axis1=3, axis2=5),
            axis=-1,
        )

    # Time lags axis
    tau = np.arange(-(n_embeddings - 1), n_embeddings) / sampling_frequency

    return tau, np.squeeze(acfs)


def raw_covariances(
    mode_covariances,
    n_embeddings,
    pca_components=None,
    zero_lag=False,
):
    """Covariance matrix of the raw channels.

    PCA and time embedding is reversed to give you to the covariance matrix
    of the raw channels.

    Parameters
    ----------
    mode_covariances : np.ndarray
        Mode covariance matrices.
    n_embeddings : int
        Number of embeddings applied to the training data.
    pca_components : np.ndarray, optional
        PCA components used for dimensionality reduction.
    zero_lag : bool, optional
        Should we return just the zero-lag elements?
        Otherwise, we return the mean over time lags.

    Returns
    -------
    raw_covs : np.ndarray
        Covariance matrix for raw channels.
    """
    # Validation
    error_message = (
        "mode_covariances must be of shape (n_channels, n_channels) or "
        + "(n_modes, n_channels, n_channels) or "
        + "(n_sessions, n_modes, n_channels, n_channels)."
    )
    mode_covariances = array_ops.validate(
        mode_covariances,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message=error_message,
    )

    # Get covariance of time embedded data
    if pca_components is not None:
        te_covs = reverse_pca(mode_covariances, pca_components)
    else:
        te_covs = mode_covariances

    if zero_lag:
        # Return the zero-lag elements only
        raw_covs = te_covs[
            :,
            :,
            n_embeddings // 2 :: n_embeddings,
            n_embeddings // 2 :: n_embeddings,
        ]

    else:
        # Return block means
        n_sessions = te_covs.shape[0]
        n_modes = te_covs.shape[1]
        n_parcels = te_covs.shape[-1] // n_embeddings

        n_parcels = te_covs.shape[-1] // n_embeddings
        blocks = te_covs.reshape(
            n_sessions,
            n_modes,
            n_parcels,
            n_embeddings,
            n_parcels,
            n_embeddings,
        )
        block_diagonal = blocks.diagonal(0, 2, 4)
        diagonal_means = block_diagonal.diagonal(0, 2, 3).mean(3)

        raw_covs = blocks.mean((3, 5))
        raw_covs[:, :, np.arange(n_parcels), np.arange(n_parcels)] = diagonal_means

    return np.squeeze(raw_covs)


def reverse_pca(covariances, pca_components):
    """Reverses the effect of PCA on covariance matrices.

    Parameters
    ----------
    covariances : np.ndarray
        Covariance matrices.
    pca_components : np.ndarray
        PCA components used for dimensionality reduction.

    Returns
    -------
    covariances : np.ndarray
        Covariance matrix of the time embedded data.
    """
    if covariances.shape[-1] != pca_components.shape[-1]:
        raise ValueError(
            "Covariance matrix and PCA components have incompatible shapes: "
            + f"covariances.shape={covariances.shape}, "
            + f"pca_components.shape={pca_components.shape}."
        )

    return pca_components @ covariances @ pca_components.T


def state_activations(state_time_course):
    """Calculate state activations from a state time course.

    Given a state time course (strictly binary), calculate the beginning and
    end of each activation of each state. Accepts a 1D or 2D array. If a 1D
    array is passed, it is assumed to be a single state time course.

    Either an array of ints or an array of :code:`bool` is accepted, but if
    :code:`int` are passed they should be explicitly 0 or 1.

    Parameters
    ----------
    state_time_course : numpy.ndarray or list of numpy.ndarray
        State time course (strictly binary).

    Returns
    -------
    slices: list of list of slice
        List containing state activations (index) in the order they occur for
        each state. This cannot necessarily be converted into an array as an
        equal number of elements in each array is not guaranteed.
    """
    # Make sure we have a list of numpy arrays
    shape_error_message = (
        "State time course must be a 1D, 2D or 3D array or list of 2D arrays."
    )
    if isinstance(state_time_course, np.ndarray):
        if state_time_course.ndim == 3 or state_time_course.dtype == object:
            state_time_course = list(state_time_course)
        elif state_time_course.ndim == 2:
            state_time_course = [state_time_course]
        elif state_time_course.ndim == 1:
            state_time_course = [state_time_course[:, np.newaxis]]
        else:
            raise ValueError(shape_error_message)
    elif isinstance(state_time_course, list):
        if not all(isinstance(stc, np.ndarray) for stc in state_time_course):
            raise ValueError(shape_error_message)
        if not all(stc.ndim == 2 for stc in state_time_course):
            raise ValueError(shape_error_message)

    # Make sure the list of arrays is of type bool
    type_error_message = (
        "State time course must be strictly binary. "
        "This can either be np.bools or np.ints with values 0 and 1."
    )
    bool_state_time_course = []
    for stc in state_time_course:
        if np.issubdtype(stc.dtype, np.integer):
            if np.all(np.isin(stc, [0, 1])):
                bool_state_time_course.append(stc.astype(bool))
        elif np.issubdtype(stc.dtype, np.bool_):
            bool_state_time_course.append(stc)
        else:
            raise ValueError(type_error_message)

    # Get the slices where each state is True
    slices = [
        [array_ops.ezclump(column) for column in stc.T]
        for stc in bool_state_time_course
    ]
    return slices


def lifetimes(state_time_course, sampling_frequency=None, squeeze=True):
    """Calculate state lifetimes from a state time course.

    Given a state time course (one-hot encoded), calculate the lifetime of each
    activation of each state.

    Parameters
    ----------
    state_time_course : numpy.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, or n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the lifetimes in seconds.
    squeeze : bool, optional
        If :code:`True`, squeeze the output to remove singleton dimensions.

    Returns
    -------
    lts : list of numpy.ndarray
        List containing an array of lifetimes in the order they occur for each
        state. This cannot necessarily be converted into an array as an equal
        number of elements in each array is not guaranteed. Shape is
        (n_sessions, n_states, n_activations) or (n_states, n_activations).
    """
    sampling_frequency = sampling_frequency or 1
    slices = state_activations(state_time_course)

    result = [
        [
            np.array([array_ops.slice_length(slice_) for slice_ in state_slices])
            / sampling_frequency
            for state_slices in session_slices
        ]
        for session_slices in slices
    ]

    if not squeeze:
        return result

    if len(result) == 1:
        result = result[0]
        if len(result) == 1:
            result = result[0]

    return result


def lifetime_statistics(state_time_course, sampling_frequency=None):
    """Calculate statistics of the lifetime distribution of each state.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the lifetimes in seconds.

    Returns
    -------
    means : np.ndarray
        Mean lifetime of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    std : np.ndarray
        Standard deviation of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    lifetimes_ = lifetimes(
        state_time_course,
        sampling_frequency=sampling_frequency,
        squeeze=False,
    )
    means = np.squeeze(array_ops.list_means(lifetimes_))
    stds = np.squeeze(array_ops.list_stds(lifetimes_))
    return means, stds


def mean_lifetimes(state_time_course, sampling_frequency=None):
    """Calculate the mean lifetime of each state.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the lifetimes in seconds.

    Returns
    -------
    mlt : np.ndarray
        Mean lifetime of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    return lifetime_statistics(state_time_course, sampling_frequency)[0]


def intervals(state_time_course, sampling_frequency=None, squeeze=True):
    """Calculate state intervals from a state time course.

    An interval is the duration between successive visits for a particular
    state.

    Parameters
    ----------
    state_time_course : list or numpy.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the intervals in seconds.
    squeeze : bool, optional
        If :code:`True`, squeeze the output to remove singleton dimensions.

    Returns
    -------
    intvs : list of numpy.ndarray
        List containing an array of intervals in the order they occur for each
        state. This cannot necessarily be converted into an array as an equal
        number of elements in each array is not guaranteed. Shape is
        (n_sessions, n_states, n_activations) or (n_states, n_activations).
    """
    sampling_frequency = sampling_frequency or 1
    slices = state_activations(state_time_course)
    result = []
    for array_slice in slices:
        r = []
        for state_slices in array_slice:
            a, b = itertools.tee(state_slices)
            next(b, None)
            state_slices_iter = zip(a, b)
            r.append(
                np.array(
                    [
                        slice_1.start - slice_0.stop
                        for slice_0, slice_1 in state_slices_iter
                    ]
                )
                / sampling_frequency
            )
        result.append(r)

    if not squeeze:
        return result

    if len(result) == 1:
        result = result[0]
        if len(result) == 1:
            result = result[0]

    return result


def interval_statistics(state_time_course, sampling_frequency=None):
    """Calculate statistics of the interval distribution of each state.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the lifetimes in seconds.

    Returns
    -------
    means : np.ndarray
        Mean interval of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    std : np.ndarray
        Standard deviation of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    intervals_ = intervals(
        state_time_course, sampling_frequency=sampling_frequency, squeeze=False
    )
    means = np.squeeze(array_ops.list_means(intervals_))
    stds = np.squeeze(array_ops.list_stds(intervals_))
    return means, stds


def mean_intervals(state_time_course, sampling_frequency=None):
    """Calculate the mean interval of each state.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If passed returns the intervals in seconds.

    Returns
    -------
    mlt : np.ndarray
        Mean interval of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    return interval_statistics(state_time_course, sampling_frequency)[0]


def fractional_occupancies(state_time_course):
    """Calculate the fractional occupancy.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).

    Returns
    -------
    fo : np.ndarray
        The fractional occupancy of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    if isinstance(state_time_course, np.ndarray):
        if state_time_course.ndim == 2:
            state_time_course = [state_time_course]
        elif state_time_course.ndim != 3:
            raise ValueError(
                "A (n_sessions, n_samples, n_states) or "
                "(n_samples, n_states) array must be passed."
            )
    fo = [np.sum(stc, axis=0) / stc.shape[0] for stc in state_time_course]
    return np.squeeze(fo)


def switching_rates(state_time_course, sampling_frequency=None):
    """Calculate the switching rate.

    This is defined as the number of state activations per second.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If :code:`None`, defaults to 1 Hz.

    Returns
    -------
    sr : np.ndarray
        The switching rate of each state. Shape is (n_sessions, n_states)
        or (n_states,).
    """
    if isinstance(state_time_course, np.ndarray):
        if state_time_course.ndim == 2:
            state_time_course = [state_time_course]
        elif state_time_course.ndim == 3:
            state_time_course = list(state_time_course)

    # Set sampling frequency
    sampling_frequency = sampling_frequency or 1

    # Loop through arrays
    sr = []
    for array in state_time_course:
        n_samples, n_states = array.shape

        # Number of activations for each state
        d = np.diff(array, axis=0)
        counts = np.array([len(d[:, i][d[:, i] == 1]) for i in range(n_states)])

        # Calculate switching rates
        sr.append(counts * sampling_frequency / n_samples)

    return np.squeeze(sr)


def mean_amplitudes(state_time_course, data):
    """Calculate mean amplitude for bursts.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    data : list or np.ndarray
        Single channel time series data (before calculating the amplitude
        envelope). Shape must be (n_sessions, n_samples, 1) or (n_samples, 1).

    Returns
    -------
    amp : np.ndarray
        Mean amplitude of the data for each state.
        Shape is (n_sessions, n_states) or (n_states,).
    """
    if isinstance(state_time_course, np.ndarray):
        if state_time_course.ndim == 2:
            state_time_course = [state_time_course]
        elif state_time_course.ndim == 3:
            state_time_course = list(state_time_course)

    if isinstance(data, np.ndarray):
        if data.ndim == 2:
            data = [data]
        elif data.ndim == 3:
            data = list(data)

    n_sessions = len(state_time_course)
    n_states = state_time_course[0].shape[1]

    # Calculate amplitude envelope of data
    data = [abs(signal.hilbert(d, axis=0)) for d in data]

    # Calculate mean amplitude envelope when each state is on
    amp = np.empty([n_sessions, n_states])
    for i in range(n_sessions):
        for j in range(n_states):
            amp[i, j] = np.mean(data[i][state_time_course[i][:, j] == 1])

    return np.squeeze(amp)


def fano_factor(state_time_course, window_lengths, sampling_frequency=1.0):
    """Calculate the Fano factor.

    Parameters
    ----------
    state_time_course : list or np.ndarray
        State time course (strictly binary). Shape must be (n_sessions,
        n_samples, n_states) or (n_samples, n_states).
    window_lengths : list or np.ndarray
        Window lengths to use. Must be in samples.
    sampling_frequency : float, optional
        Sampling frequency in Hz.

    Returns
    -------
    F : list of np.ndarray
        Fano factor. Shape is (n_sessions, n_window_lengths, n_states) or
        (n_window_lengths, n_states).
    """
    if isinstance(state_time_course, np.ndarray):
        state_time_course = [state_time_course]

    # Loop through arrays
    F = []
    for array in state_time_course:
        n_samples = array.shape[0]
        n_states = array.shape[1]
        F.append([])

        # Loop through window lengths
        for window_length in window_lengths:
            w = int(window_length * sampling_frequency)
            n_windows = n_samples // w
            tc = array[: n_windows * w]
            tc = tc.reshape(n_windows, w, n_states)

            # Loop through windows
            counts = []
            for window in tc:
                # Number of activations
                d = np.diff(window, axis=0)
                c = []
                for i in range(n_states):
                    c.append(len(d[:, i][d[:, i] == 1]))
                counts.append(c)

            # Calculate Fano factor
            counts = np.array(counts)
            F[-1].append(np.std(counts, axis=0) ** 2 / np.mean(counts, axis=0))

    return np.squeeze(F)


def calc_trans_prob_matrix(state_time_course, n_states=None):
    """Calculate session-specific transition probability matrices.

    Parameters
    ----------
    state_time_course : list of np.ndarray or np.ndarray
        State time courses. Shape must be (n_sessions, n_samples, n_states)
        or (n_samples, n_states).
    n_states : int, optional
        Number of states.

    Returns
    -------
    trans_prob : np.ndarray
        Session-specific transition probability matrices. Shape is (n_sessions,
        n_states, n_states).
    """
    if isinstance(state_time_course, np.ndarray):
        state_time_course = [state_time_course]
    trans_prob = []
    for stc in state_time_course:
        stc_argmax = stc.argmax(axis=1)
        vals, counts = np.unique(
            stc_argmax[np.arange(2)[None, :] + np.arange(len(stc_argmax) - 1)[:, None]],
            axis=0,
            return_counts=True,
        )
        if n_states is None:
            n_states = stc_argmax.max() + 1
        tp = np.zeros((n_states, n_states))
        tp[vals[:, 0], vals[:, 1]] = counts
        with np.errstate(divide="ignore", invalid="ignore"):
            tp /= tp.sum(axis=1)[:, None]
        trans_prob.append(np.nan_to_num(tp))
    return np.squeeze(trans_prob)


def simple_moving_average(data, window_length, step_size):
    """Calculate imple moving average.

    This function can be used to calculate a sliding window fractional occupancy
    from a state time course. This was done in `Baker et al. (2014)
    <https://elifesciences.org/articles/01867>`_.

    Parameters
    ----------
    data : np.ndarray
        Time series data. Shape must be (n_samples, n_channels).
    window_length : int
        Number of data points in a window.
    step_size : int
        Step size for shifting the window.

    Returns
    -------
    mov_avg : np.ndarray
        Mean for each window.
    """
    # Get number of samples and modes
    n_samples = data.shape[0]
    n_modes = data.shape[1]

    # Pad the data
    data = np.pad(data, window_length // 2)[
        :,
        window_length // 2 : window_length // 2 + n_modes,
    ]

    # Define indices of time points to calculate a moving average
    time_idx = range(0, n_samples, step_size)
    n_windows = n_samples // step_size

    # Preallocate an array to hold moving average values
    mov_avg = np.empty([n_windows, n_modes], dtype=np.float32)

    # Compute simple moving average
    for n in range(n_windows):
        j = time_idx[n]
        mov_window = data[j : j + window_length]
        mov_avg[n] = np.mean(mov_window, axis=0)

    return mov_avg


def partial_covariances(data, alpha):
    r"""Calculate partial covariances.

    Returns the multiple regression parameters estimates of the state/mode time
    courses regressed onto the data from each channel. The regression parameters
    are referred to as 'partial covariances'.

    We fit the regression:

    .. math::
        Y_i = X \beta_i + \epsilon

    where:

    - :math:`Y_i` is (n_samples, 1) the data amplitude/envelope/power/absolute
      time course at channel :math:`i`.
    - :math:`X` is (n_samples, n_states) matrix of the variance normalised
      state/mode time courses.
    - :math:`\beta_i` is an (n_states, 1) vector of multiple regression
      parameters for channel :math:`i`.
    - :math:`\epsilon` is the error.

    Parameters
    ----------
    data : np.ndarray or list of np.ndarray
        Training data for each array. Shape is (n_sessions, n_samples,
        n_channels) or (n_samples, n_channels).
    alpha : np.ndarray or list of np.ndarray
        State/mode time courses for each array. Shape is (n_sessions,
        n_samples, n_states) or (n_samples, n_states).

    Returns
    -------
    partial_covariances : np.ndarray
        Matrix of partial covariance (multiple regression parameter estimates,
        :math:`\beta`). Shape is (n_states, n_channels).

    Note
    ----

    - The regression is done separately for each channel.
    - State/mode time courses are variance normalized so that all amplitude
      info goes into the partial covariances, :math:`\beta_i`.
    """
    if type(data) != type(alpha):
        raise ValueError(
            "data and alpha must be the same type: numpy arrays or lists of "
            "numpy arrays."
        )
    if isinstance(data, np.ndarray):
        data = [data]
        alpha = [alpha]
    for i in range(len(data)):
        if data[i].shape[0] != alpha[i].shape[0]:
            raise ValueError("Difference number of samples in data and alpha.")

    pcovs = []
    for X, a in zip(data, alpha):
        # Variance normalise state/mode time courses
        a_normed = a / np.std(a, axis=0, keepdims=True)

        # Do multiple regression of alpha onto data
        pcovs.append(np.linalg.pinv(a_normed) @ X)

    return np.squeeze(pcovs)


def hmm_dual_estimation(data, alpha, zero_mean=False, eps=1e-5, n_jobs=1):
    """HMM dual estimation of observation model parameters.

    Parameters
    ----------
    data : np.ndarray or list of np.ndarray
        Time series data. Shape must be (n_samples, n_channels)
        or (n_subjects, n_samples, n_channels).
    alpha : np.ndarray or list of np.ndarray
        State probabilities. Shape must be (n_samples, n_states)
        or (n_subjects, n_samples, n_states).
    zero_mean : bool, optional
        Should we force the state means to be zero?
    eps : float, optional
        Small value to add to the diagonal of each state covariance.
    n_jobs : int, optional
        Number of jobs to run in parallel. If set as None, the function
        will run sequentially.

    Returns
    -------
    means : np.ndarray or list of np.ndarray
        State means. Shape is (n_states, n_channels) or
        (n_subjects, n_states, n_channels).
    covariances : np.ndarray or list of np.ndarray
        State covariances. Shape is (n_states, n_channels, n_channels)
        or (n_subjects, n_states, n_channels, n_channels).
    """

    # Validation
    if (isinstance(data, list) != isinstance(alpha, list)) or (
        isinstance(data, np.ndarray) != isinstance(alpha, np.ndarray)
    ):
        raise ValueError(
            f"data is type {type(data)} and alpha is type "
            f"{type(alpha)}. They must both be lists or numpy arrays."
        )

    if isinstance(data, np.ndarray):
        if alpha.shape[0] != data.shape[0]:
            raise ValueError("data and alpha must have the same number of samples.")

        if data.ndim == 2:
            data = [data]
            alpha = [alpha]

    if len(data) != len(alpha):
        raise ValueError(
            "A different number of arrays has been passed for "
            f"data and alpha: len(data)={len(data)}, "
            f"len(alpha)={len(alpha)}."
        )

    # Check the number of samples in data and alpha
    for i in range(len(alpha)):
        if alpha[i].shape[0] != data[i].shape[0]:
            raise ValueError(
                "items in data and alpha must have the same number of samples."
            )

    n_states = alpha[0].shape[1]
    n_channels = data[0].shape[1]

    # Helper function
    def _calc(a, x):
        sum_a = np.sum(a, axis=0)

        n_samples = x.shape[0]
        memory_threshold = 1e8  # threshold for memory usage
        seq_length = max(int(memory_threshold // (n_channels**2)), 1)

        m = np.zeros([n_states, n_channels])
        if not zero_mean:
            for i in range(n_states):
                m[i] = np.sum(x * a[:, i, None], axis=0) / sum_a[i]

        c = np.zeros([n_states, n_channels, n_channels])
        for i in range(n_states):
            if x.size <= memory_threshold:
                d = x - m[i]
                c[i] = (
                    np.sum(d[:, :, None] * d[:, None, :] * a[:, i, None, None], axis=0)
                    / sum_a[i]
                )
            else:
                # If the data is too large, calculate in chunks to avoid memory overflow.
                for start in range(0, n_samples, seq_length):
                    end = min(start + seq_length, n_samples)
                    d = x[start:end] - m[i]
                    c[i] += np.sum(
                        d[:, :, None] * d[:, None, :] * a[start:end, i, None, None],
                        axis=0,
                    )
                c[i] /= sum_a[i]
            c[i] += eps * np.eye(n_channels)

        return m, c

    if n_jobs is None:
        results = [_calc(a, x) for a, x in zip(alpha, data)]
    else:
        # Calculate in parallel
        results = pqdm(
            array=zip(alpha, data),
            function=_calc,
            n_jobs=n_jobs,
            desc="Dual estimation",
            argument_type="args",
            total=len(data),
        )

    # Unpack results
    means = []
    covariances = []
    for result in results:
        m, c = result
        means.append(m)
        covariances.append(c)

    return np.squeeze(means), np.squeeze(covariances)


def hmm_features(
    data,
    alpha,
    sampling_frequency=None,
    zero_mean=False,
    eps=1e-5,
    use_partial=False,
    n_jobs=1,
):
    """Dual estimation of HMM features.

    Parameters
    ----------
    data : np.ndarray or list of np.ndarray
        Prepared data. Shape must be (n_samples, n_channels)
        or (n_subjects, n_samples, n_channels).
    alpha : np.ndarray or list of np.ndarray
        State probabilities. Shape must be (n_samples, n_states)
        or (n_subjects, n_samples, n_states).
    sampling_frequency : float, optional
        Sampling frequency in Hz. If not passed, summary statistics
        are unitless.
    zero_mean : bool, optional
        Should we force the state means to be zero?
    eps : float, optional
        Small value to add to the diagonal of each state covariance.
    use_partial : bool, optional
        Should we use the partial state correlation matrix rather than
        the full state covariance matrix?
    n_jobs : int, optional
        Number of jobs to run in parallel.

    Returns
    -------
    features : np.ndarray
        HMM features. Shape is (n_subjects, n_features).
    """
    from osl_dynamics.inference.modes import argmax_time_courses

    # Validation
    if (isinstance(data, list) != isinstance(alpha, list)) or (
        isinstance(data, np.ndarray) != isinstance(alpha, np.ndarray)
    ):
        raise ValueError(
            f"data is type {type(data)} and alpha is type "
            f"{type(alpha)}. They must both be lists or numpy arrays."
        )

    if isinstance(data, np.ndarray):
        if alpha.shape[0] != data.shape[0]:
            raise ValueError("data and alpha must have the same number of samples.")

        if data.ndim == 2:
            data = [data]
            alpha = [alpha]

    if len(data) != len(alpha):
        raise ValueError(
            "A different number of arrays has been passed for "
            f"data and alpha: len(data)={len(data)}, "
            f"len(alpha)={len(alpha)}."
        )

    def _calc(a, x):
        # Summary statistics for dynamics
        stc = argmax_time_courses(a)
        fo = fractional_occupancies(stc)
        lt = mean_lifetimes(stc, sampling_frequency=sampling_frequency)
        intv = mean_intervals(stc, sampling_frequency=sampling_frequency)
        sr = switching_rates(stc, sampling_frequency=sampling_frequency)
        sum_stats = np.concatenate([fo, lt, intv, sr], axis=-1)

        # Transition probabilities
        trans_prob = calc_trans_prob_matrix(stc, n_states=stc.shape[-1])
        trans_prob = trans_prob.flatten()

        # Observation model parameters
        m, c = hmm_dual_estimation(x, a, zero_mean=zero_mean, eps=eps, n_jobs=None)
        i, j = np.triu_indices(c.shape[-1])
        if use_partial:
            c = array_ops.cov2partialcorr(c)
        c = c[..., i, j]
        if zero_mean:
            obs_mod = c
        else:
            obs_mod = np.concatenate([m, c], axis=-1)
        obs_mod = obs_mod.reshape(-1)

        # Combine features
        return np.concatenate([sum_stats, trans_prob, obs_mod])

    # Calculate in parallel
    features = pqdm(
        array=zip(alpha, data),
        function=_calc,
        n_jobs=n_jobs,
        desc="Calculating HMM features",
        argument_type="args",
        total=len(data),
    )

    return np.squeeze(features)



"""Functions to perform regression analysis."""

import logging

import numpy as np
from sklearn.linear_model import LinearRegression

_logger = logging.getLogger("osl-dynamics")


def linear(X, y, fit_intercept, normalize=False, log_message=False):
    """Wrapper for `sklearn.linear_model.LinearRegression \
    <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model\
    .LinearRegression.html>`_.

    Parameters
    ----------
    X : np.ndarray
        Regressors, should be a 2D array (n_targets, n_regressors).
    y : np.ndarray
        Targets. Should be a 2D array: (n_targets, n_features).
        If a higher dimension array is passed, the extra dimensions
        are concatenated.
    fit_intercept : bool
        Should we fit an intercept?
    normalize : bool, optional
        Should we z-transform the regressors?
    log_message : bool, optional
        Should we log a message?

    Returns
    -------
    coefs : np.ndarray
        Regression coefficients. 2D array or higher dimensionality:
        (n_regressors, n_features).
    intercept : np.ndarray
        Regression intercept. 1D array or higher dimensionality:
        (n_features,). Returned if :code:`fit_intercept=True`.
    """
    if log_message:
        _logger.info("Fitting linear regression")

    # Reshape in case non 2D matrices were passed
    original_shape = y.shape
    new_shape = [X.shape[1]] + list(original_shape[1:])
    y = y.reshape(original_shape[0], -1)

    # Normalise the regressors
    if normalize:
        X -= np.mean(X, axis=0)
        X /= np.std(X, axis=0)

    if y.dtype == np.complex64 or y.dtype == np.complex_:
        # Fit two linear regressions:
        # One for the real part
        reg = LinearRegression(fit_intercept=fit_intercept, n_jobs=-1)
        reg.fit(X, y.real)
        coefs_real = reg.coef_.T.reshape(new_shape)
        if fit_intercept:
            intercept_real = reg.intercept_.reshape(new_shape[1:])

        # Another for the imaginary part
        reg = LinearRegression(fit_intercept=fit_intercept, n_jobs=-1)
        reg.fit(X, y.imag)
        coefs_imag = reg.coef_.T.reshape(new_shape)
        if fit_intercept:
            intercept_imag = reg.intercept_.reshape(new_shape[1:])

        # Regression parameters
        coefs = coefs_real + 1j * coefs_imag
        if fit_intercept:
            intercept = intercept_real + 1j * intercept_imag

    else:
        # Only need to fit one linear regression
        reg = LinearRegression(fit_intercept=fit_intercept, n_jobs=-1)
        reg.fit(X, y)

        # Regression parameters
        coefs = reg.coef_.T.reshape(new_shape)
        if fit_intercept:
            intercept = reg.intercept_.reshape(new_shape[1:])

    if fit_intercept:
        return coefs, intercept
    else:
        return coefs
